use std::{
    io::Read,
    sync::{Arc, RwLock},
    fs::File,
    path::Path,
};

use trie::Trie;

use crate::util::string::minimize_whitespace;

#[derive(Clone)]
pub enum Source {
    Raw(Vec<u8>),
    String(String),
    File(std::path::PathBuf),
    MarkdownFile(std::path::PathBuf),
}

#[derive(Clone)]
pub struct CacheData {
    pub mime:String,
    pub data:Vec<u8>,
}

struct WebCacheData {
    objects:Trie<CacheData>,
}

#[derive(Clone)]
pub struct WebCache {
    data:Arc<RwLock<WebCacheData>>,
}
impl WebCache {
    pub fn new() -> Self
    {
        Self {
            data:Arc::new(RwLock::new(WebCacheData {
                objects:Trie::new(),
            })),
        }
    }

    pub fn fetch(&self, object:&str) -> Option<CacheData>
    {
        match self.data.read() {
            Ok(reader) => {
                match reader.objects.get(object.as_bytes()) {
                    Some(data) => Some(data.clone()),
                    None => None,
                }
            }
            Err(_) => None,
        }
    }

    pub fn cache(&self, mime:&str, object:&str, sources:&[Source]) -> Result<(),()>
    {
        match self.data.write() {
            Ok(mut writer) => {
                let mut data = Vec::new();

                for source in sources {
                    match source {
                        Source::Raw(raw) => {
                            data.append(&mut raw.clone());
                        }
                        Source::String(text) => {
                            data.append(&mut text.as_bytes().to_vec());
                        }
                        Source::File(path) => {
                            let mut file_data = Vec::new();
                            let mut file = File::open(path).map_err(|_| ())?;
                            file.read_to_end(&mut file_data).ok();
                            data.append(&mut file_data);
                        }
                        Source::MarkdownFile(path) => {
                            let mut text = markdown::file_to_html(&path.as_ref()).map_err(|_| ())?;
                            text = text.trim().lines().map(|line| line.trim().to_string()).collect();
                            data.append(&mut text.as_bytes().to_vec());
                        }
                    }
                }

                writer.objects.set(object.as_bytes(), CacheData {
                    mime:mime.to_string(),
                    data,
                });
                Ok(())
            },
            Err(_) => Err(()),
        }
    }

    pub fn cache_whitespace_minimize(&self, object:&str) -> Result<(),()>
    {
        match self.data.write() {
            Ok(mut writer) => {
                if let Some(data) = writer.objects.get_mut(object.as_bytes()) {
                    data.data = minimize_whitespace(&String::from_utf8(data.data.clone()).unwrap()).as_bytes().to_vec();
                    Ok(())
                } else { Err(()) }
            },
            Err(_) => Err(()),
        }
    }


    pub fn raw(data:Vec<u8>) -> Source { Source::Raw(data) }
    pub fn string(text:&str) -> Source { Source::String(text.to_string()) }
    pub fn file<P:AsRef<Path>>(path:P) -> Source { Source::File((path.as_ref() as &Path).to_path_buf())}
    pub fn markdown<P:AsRef<Path>>(path:P) -> Source { Source::MarkdownFile((path.as_ref() as &Path).to_path_buf())}
}
